// VinciaMatching.cc is a part of the VINCIA plug-in to the PYTHIA 8 
// generator.
// VINCIA is licenced under the GNU GPL version 2.
// Copyright (C) 2013 P. Skands.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// Function definitions (not found in the header) for the Matching class. 
#include "VinciaMatching.h"

using namespace Pythia8;

namespace Vincia {

  //***********************************************************************
  // init() : 
  // Initialize Matching interface
  //***********************************************************************

  bool Matching::init(int nVariationsIn) {

    // Check if pointers initialized
    if (! isInitPtr) {
      printOut("VM::init","Cannot initialize, pointers not set.");
      return false;
    }
    
    // Innocent until proven guilty
    isInit = true;

    // Set number of uncertainty variations (at least one, for USR)
    nVariations = max(1,nVariationsIn);

    // Initialize Settings
    verbose          = settingsPtr->mode("Vincia:verbose");
    uncertaintyBands = settingsPtr->flag("Vincia:uncertaintyBands");
    
    // Initialize Matching Settings 
    matchingFullColor = settingsPtr->flag("Vincia:matchingFullColor");
    matchingNLO       = settingsPtr->flag("Vincia:matchingNLO");

    // Initialize Shower Settings (needed to compute matching coefficients)
    evolutionType   = settingsPtr->mode("Vincia:evolutionType");
    orderingMode    = settingsPtr->mode("Vincia:orderingMode");
    sectorShower    = settingsPtr->flag("Vincia:sectorShower");
    showerAlphaScmw = settingsPtr->flag("Vincia:alphaScmw");    
    pTnormalization = settingsPtr->parm("Vincia:pTnormalization");
    mDnormalization = settingsPtr->parm("Vincia:mDnormalization");

    // Check and initialize MG interface 
    isOnMG = settingsPtr->flag("Vincia:MGInterface");
    if (isOnMG) isOnMG = mgInterfacePtr->init();
    
    // Check and initialize BH interface
    isOnBH = false;

    return isInit;
  }
	
  //***********************************************************************
  // Main functions for VINCIA matching
  //***********************************************************************

  // *********************************************************************
  // isMatched() : return true if matching exists at this order
  // *********************************************************************

  bool Matching::isMatched(int nIn, vector<int> idPartons, bool isPolarized) {
    
    // Sanity check
    if (nIn != 1) return false;
    if (idPartons.size() <= 2) return false;

    // Return value false by default
    bool gotit = false;    

    // Data about resonance
    int colType  = particleDataPtr->colType(idPartons[0]);
    int chgType  = abs(particleDataPtr->chargeType(idPartons[0]));
    int spinType = particleDataPtr->spinType(idPartons[0]); 

    // Data about final state
    int nFS = idPartons.size() - nIn;
    int nG(0), nQ(0), nMisc(0);
    bool isMassive = false;
    for (int i=nIn; i<int(idPartons.size()); ++i) {
      if (idPartons[i] == 21) ++nG;
      else if (idPartons[i] <= 8) ++nQ;
      else ++nMisc;
      if (vinComPtr->isMassive(idPartons[i])) isMassive = true; 
    }

    // Helicity-dependence is switched off for massive particles
    if (isMassive) isPolarized = false;

    // So far matching only for pure gluon+quark final states
    if (nMisc >= 1) return false;

    // -------------------------------------------------------------------
    // Matching: with MG interface
    // -------------------------------------------------------------------
    if (isOnMG) {

      // Colorless resonances
      if (colType == 0) {

	// Scalar resonances (also used for undertermined spins)
	if (spinType <= 1) {	  

	  // Neutral (H0-like)
	  if (chgType == 0) {

            // H->qq, up to six partons
            if (nQ >= 2 && nFS <= 6) gotit = true; 

            // Pure gluons, up to H -> 5 gluons
            else if (nQ == 0 && nG <= 5) gotit = true;

          }

          // Charged (H+-like): only for helicity-dependent matching
	  else if (chgType == 3 && isPolarized) {

            // H->qq, up to six partons
            if (nQ >= 2 && nFS <= 6) gotit = true;
          }
          
        }
        
	// Vector resonances 
	else if (spinType == 3) {

	  // Neutral (Z-like)
	  if (chgType == 0) {
            
            // Z->qq, up to 6 jets
            if (nQ >= 2 && nFS <= 6) gotit = true;
	  
          }

	  // Charged (W-like): only for helicity-dependent matching
	  else if (chgType == 3 && isPolarized) {
            
            // W->qq, up to 6 jets
            if (nQ >= 2 && nFS <= 6) gotit = true;
            
          }

        }

      }

    }

    // -------------------------------------------------------------------
    // Matching: with BlackHat interface
    // -------------------------------------------------------------------
    if (!gotit && isOnBH) {

      // So far no BlackHat matrix elements implemented
      gotit = false;

    }

    return gotit;

  } 

  // *********************************************************************
  // isMatchedNLO() : return true if NLO matching exists at this order
  // *********************************************************************
  
  bool Matching::isMatchedNLO(int nIn, vector<int> idParticles) {
    
    // Default is we don't have it
    bool gotIt = false;
    
    // Need at least 3 particles to be involved
    if (idParticles.size() <= 2) return false;
    
    // NLO matching only for decays so far
    if (nIn >= 2) return false;
    
    // Z decays
    if (idParticles[0] == 23) {
      
      // Z: NLO matching exists for Z->2,3 partons
      if (idParticles.size() == 3) gotIt = true;
      else if (idParticles.size() == 4) gotIt = true;
      
    }

    return gotIt;
    
  }

  // *********************************************************************
  // polarize() : use helicity matrix elements to polarize a list of 
  // partons 
  // *********************************************************************

  bool Matching::polarize(int nIn, vector<Particle*> partons) {

    // So far, we only know how to polarize resonance decays
    if (nIn != 1 || partons.size() < 3) return false;

    // Resonance decays
    if (nIn == 1) {

      // Identify unpolarized partonic state and check color ordering
      vector<int> idPartons;
      vector<Vec4> pPartons;
      vector<double> polPartons;
      for (int i=0; i<int(partons.size()); ++i) {
	idPartons.push_back(partons[i]->id());
	pPartons.push_back(partons[i]->p());
	polPartons.push_back(9.);
	int acol = partons[i]->acol();
	if (i >= 1 && acol > 0) { 
	  if (partons[i-1]->col() != acol) {
	    if (verbose >= 3) 
	      printOut("VM::polarize",
		       "failed to polarize state. Not color ordered.");
	    return false;
	  }
	}
      }
      // Allow decaying resonance to be polarized
      polPartons[0] = partons[0]->pol();

      // Check if we have the relevant matrix element
      if (!isMatched(1,idPartons,true)) {
	if (verbose >= 6) 
	  printOut("VM::polarize",
		   "failed to polarize state. Unknown matrix element.");
	return false;      
      }
      
      // Get the polarization-summed ME
      double sumAll = mgInterfacePtr->ME2(1,idPartons,pPartons,polPartons);

      // Generate random number
      double sumR  = rndmPtr->flat() * sumAll;

      // Now loop over final-state polarizations until sum > sumR
      int npols = (int)pow(2.,(int)idPartons.size() - nIn);
      double sumNow = 0;
      for (int i = 0; i < npols; i++) {
	// Remove all but decaying resonance from polPartons,
	// then set this polarization state
	polPartons.resize(1);
	for (int j = 1; j <= (int)idPartons.size() - nIn; ++j) {	  
	  if (i % (int)pow(2.,j) < (int)pow(2.,j-1)) polPartons.push_back(-1.);
	  else polPartons.push_back(1.);
	}
	sumNow += mgInterfacePtr->ME2(1,idPartons,pPartons,polPartons);
	// Verbose output
	if (verbose >= 6) {
	  cout<<"(VM::polarize) Tested "<<polPartons[0]<<" -> ";
	  for (int j=1; j<= int(idPartons.size()) - 1; ++j) {
	    cout<<polPartons[j]<<" ";
	  }
	  cout<<" sumNow = "<<sumNow<<endl;	
	}
	// Check if this is the selected polarization
	if (sumNow >= sumR) break;
      }
      
      if (verbose >= 5) {
	cout<<"(VM::polarize) Chose FS polarizations = ";
	for (int j=1; j<= int(idPartons.size()) - 1; ++j) {
	  cout<<polPartons[j]<<" ";
	}
	cout<<endl;
      }

      // Save chosen polarizations to incoming parton list
      for (int i=nIn; i<int(partons.size()); ++i) {
	partons[i]->pol(polPartons[i]);
      }

      // All is well
      return true;

    }

    // Should only arrive here if nIn != 1: failed.
    return false;
    
  }

  // Wrapper to operate directly on particles in an event record
  bool Matching::polarize(Event& event, int nIn, vector<int> iPartons) {
    
    // Sanity check
    if (nIn <= 0) return false;

    vector<Particle*> partons;
    for (int i=0; i<int(iPartons.size()); ++i) {
      int iNow = iPartons[i];
      if (iNow <= 0 || iNow >= event.size()) return false;
      else partons.push_back(&event[iNow]);
    }

    return polarize(nIn,partons);
  }

  // *********************************************************************
  // getME() : Return a tree-level matrix element (squared)
  // INPUT: 
  //      nIn : 1 (for decay) or 2 (for 2->2 process)
  //      id  : list of particle PDG codes
  //            incoming first, then outgoing ones
  //      p   : same for momenta
  //      pol : same for polarizations
  //   Note: the particle list is assumed to be color ordered
  //
  // RETURN VALUE:
  //      -1 if no matrix element found, otherwise |M|^2
  // *********************************************************************

  double Matching::getME(int nIn, vector<int> id, vector<Vec4> p, 
			 vector<double> pol) {

    // Initialize to -1 return value
    double ME2 = -1.0;

    // First try MadGraph
    if (isOnMG) ME2 = mgInterfacePtr->ME2(nIn, id, p, pol);

    // Then try BlackHat 
    if (ME2 <= 0.0 && isOnBH) {
      // This part so far purely dummy. 
    }

    return ME2;

  }

  // Wrapper to use list of particles as input
  double Matching::getME(int nIn, vector<Particle*> particlePtrs) {

    vector<int> id;
    vector<Vec4> p;
    vector<double> pol;
    for (int i=0; i<int(particlePtrs.size()); ++i) {
      id.push_back(particlePtrs[i]->id());
      p.push_back(particlePtrs[i]->p());
      pol.push_back(particlePtrs[i]->pol());
    }
    
    return getME(nIn, id, p, pol);

  }

  // *********************************************************************
  // getPME() : Return a matching coefficient
  // INPUT: 
  //      nIn : 1 (for decay) or 2 (for 2->2 process)
  //      particlePtrs : vector containing pointers to particles
  //           first incoming particle(s), then outgoing ones
  //   Note   : the particle list is assumed to be color-ordered
  // OUTPUT: 
  //      PME : vector whose [0] component is the matching coefficient
  //            for the current shower settings, with the additional ones
  //            corresponding to uncertainty variations (if switched on).
  //              = |M_n|^2 / (Sum_i A^phys_i |M_n-1|^2)
  //      PME[0]   : reweighting factor with USR antennae
  //      PME[1]   : reweighting factor with DEF antennae 
  //      PME[2,3] : no difference at tree level (muR-variation)
  //      PME[4]   : reweighting factor with MAX antennae
  //      PME[5]   : reweighting factor with MIN antennae
  //      PME[6,7] : no difference at tree level (NLO K-factor variation)
  //      PME[8]   : reweighting factor with stronger (quartic) ordering
  //      PME[9]   : reweighting factor with smooth mD ordering
  //      PME[10]  : reweighting factor with all emission color factors = CA 
  //      PME[11]  : reweighting factor with all emission color factors = CF
  // *********************************************************************

  bool Matching::getPME(int nIn, vector<Particle*> particlePtrs, 
			vector<double>& PME) {

    // Resize PME.
    PME.resize(0);
    bool done = false;

    // Matching using MadGraph
    if (! isOnMG) return false;

    // Create vectors of PDG ID codes, momenta, and polarizations
    vector<int> id;
    vector<Vec4> p;
    vector<double> pol;
    
    //Set all vectors to be Z bosons and all scalars to be H bosons
    // Decays: determine if we know the relevant Lorentz/Color structure
    int id0 = 0;
    // Color singlets
    if (particleDataPtr->colType(particlePtrs[0]->id()) == 0) {
      // Spin-1 (e.g., W, Z, Z', ...)
      if (particleDataPtr->spinType(particlePtrs[0]->id()) == 3) id0 = 23;
      // Spin-0 (e.g., h, h+, ...)
      if (particleDataPtr->spinType(particlePtrs[0]->id()) == 1) id0 = 25;
    }

    // If not known, return.
    if (id0 == 0) return -1;
    id.push_back(id0);
    p.push_back(particlePtrs[0]->p());
    pol.push_back(particlePtrs[0]->pol());
    
    for (int i=1; i<int(particlePtrs.size()); ++i) {
      id.push_back(particlePtrs[i]->id());
      p.push_back(particlePtrs[i]->p());
      pol.push_back(particlePtrs[i]->pol());
    }

    /*
    cout<<"ME : pols = ";
    for (unsigned int i=0; i<id.size(); ++i) cout<<" "<<pol[i];
    cout<<endl;
    */

    // Get color-ordered M2 (= LC + ...) for this n-parton state
    double Mn2     = getME(1, id, p, pol);

    // mgInterface returns -1 if not existing. 
    if (abs(Mn2+1) < TINY) return false;

    // Initialize (n-1)->n vector of approximations
    vector<double> Mn2Appr(0);
    for (int iVar=0; iVar<nVariations; iVar++) {
      Mn2Appr.push_back(0.);
    }

    // SECTOR showers: find appropriate sector    
    // Loop over possible sectors, start with first one
    // Sector decomposition variables (see [LVS11]):
    //     10: QS2 = ARIADNE pT2 = sIK * yij * yjk 
    //      8: QS2 = sIK * yjk * sqrt(yij) (for jk collinear g -> qbar q)
    //      9: QS2 = sIK * yij * sqrt(yjk) (for ij collinear g -> qbar q)
    int irMin    = 2;
    if (sectorShower) {
      // Start with first clustering: ir = 2:
      double q2Min(0.);
      if (id[irMin] == 21) q2Min = RPtr->Q2E(p[irMin-1],p[irMin],p[irMin+1],10);
      else if (id[irMin] < 0) 
	q2Min = RPtr->Q2E(p[irMin-1],p[irMin],p[irMin+1],8);
      else 
	q2Min = RPtr->Q2E(p[irMin-1],p[irMin],p[irMin+1],9);
      // Loop over additional clusterings, ir >= 3:
      for (int ir=3; ir<int(p.size())-1; ++ir) {
	double q2(0.);
	if (id[ir] == 21) q2 = RPtr->Q2E(p[ir-1],p[ir],p[ir+1],10);
	else if (id[ir] < 0)
	  q2 = RPtr->Q2E(p[ir-1],p[ir],p[ir+1],8);
	else 
	  q2 = RPtr->Q2E(p[ir-1],p[ir],p[ir+1],9);
	if (q2 < q2Min) {
	  q2Min = q2;
	  irMin = ir;
	}
      }    
      if (verbose >= 8) 
	cout<<"VM: chose sector "<<irMin<<" with clusterer id "
	    <<id[irMin]<<endl;
    }

    // Loop over possible emitter antennae
    for (int iClu=0; iClu<int(id.size())-3; iClu++) {
      
      // Partons to be clustered
      int ir    = 2+iClu;

      // SECTOR showers: skip all clusterings except correct sector
      if (sectorShower && ir != irMin) continue;

      // Verbose
      if (verbose >= 8) printOut("VM::getPME","Doing clustering no. "
                                 +num2str(iClu));
      
      // Compute basic quantities
      double ma = ( vinComPtr->isMassive(id[ir-1]) ? p[ir-1].mCalc() : 0.0 ) ; 
      double mr = ( vinComPtr->isMassive(id[ir])   ? p[ir].mCalc() : 0.0 );
      double mb = ( vinComPtr->isMassive(id[ir+1]) ? p[ir+1].mCalc() : 0.0 );
      double sP  = m2(p[ir-1]+p[ir]+p[ir+1]);
      double mP  = sqrt(sP);
      double mMua = ma / mP;
      double mMur = mr / mP; 
      double mMub = mb / mP;
      
      //Check to see if any particles are unpolarized
      //If so, then we match to fully unpolarized matrix elements
      //Ignore initial decaying particle
      bool isPol = true;
      for (int ipol = 1; ipol < (int)pol.size(); ipol++) {
        if (pol[ipol] > 8.0) {
          isPol = false;
          break;
        }
      }

      // Reconstruct previous state
      vector<int> idClu;
      vector<Vec4> pClu;
      vector< vector<int> > polClu;
      int antIndx=-1;
      bool swapAB=false;
      if (isPol) {
	// Determine all polarized clusterings as recorded in polClu
	// Note: if any of the particles are unpolarized -> unpolarized
	vinCluPtr->getCluPol(id,p,ir-1,ir,ir+1,idClu,pClu,antIndx,swapAB,
			     pol,polClu);
      } else {
	vinCluPtr->getClu(id,p,ir-1,ir,ir+1,idClu,pClu,antIndx,swapAB);
      }

      // Invariant mass of parent antenna and daughter invariants
      double yar = (m2(p[ir-1],p[ir])-pow2(ma)-pow2(mr))/sP;
      double yrb = (m2(p[ir],p[ir+1])-pow2(mr)-pow2(mb))/sP;
      if (swapAB) {
	double ytmp = yar;
	yar = yrb;
	yrb = ytmp;
        // MR 17.1.2011: have to swap masses as well
        double mMutmp = mMua;
        mMua = mMub;
        mMub = mMutmp;
      }

      // Check for vanishing invariants
      if (yar < TINY || yrb < TINY) {
	printOut("VM::getPME","Warning: small invariant(s): y = "
		 +num2str(yar)+","+num2str(yrb)
		 +" sParent ="+num2str(sP));
	if (verbose >= 3) {
	  cout<<"Momentum Listing:"<<endl;
	  cout<<p[ir-1]<<p[ir]<<p[ir+1];
	  cout<<"Invariants: "<<m2(p[ir-1]+p[ir])<<" "<<m2(p[ir]+p[ir+1])<<endl;
	}
      }
      
      // Compute Pimp factor (only used for global showers) 
      //  : orderingMode <= 1 : Pimp = 1
      //  : orderingMode == 2 : Q2E/(Q2Ehat+Q2E) 
      double Pimp         = 1.0;
      // Additional Pimp factor variations used for uncertainty estimates
      double PimpStronger = 1.0;
      double PimpMD       = 1.0;

      // Only apply if at least 3 partons in final state
      if ( !sectorShower && orderingMode >= 2 && int(pClu.size()-2) > nIn) { 

	// QE, pT, and mD Values for current clustering	
	double QE2New, mD2New;
	// If this is an anti fermion -> always use m2rb for g->qqbar
	if (id[ir] < 0) {
	    QE2New = m2(p[ir],p[ir+1]);
	    mD2New = QE2New;
	}
	// Else if this is a fermion -> always use m2ar for g->qqbar
	if (id[ir] < 0) {
	    QE2New = m2(p[ir-1],p[ir]);
	    mD2New = QE2New;
	}
	// Else use QE, pT, mD definitions for g emission
	else {
	  QE2New = RPtr->yE(yar,yrb)*sP;
	  mD2New = RPtr->yE(yar,yrb,2)*sP;
	}

	// Initial values for Hat variables characterizing the reclustered
	// parton state
	double QE2Hat=m2(pClu[1],pClu[2],pClu[3]);
	double mD2Hat=QE2Hat;
	// Loop over possible clusterings, find smallest Q for each measure
	for (int jClu=0; jClu<int(pClu.size())-3; jClu++) {
	  int jaHat=jClu+1;
	  int jrHat=jClu+2;
	  int jbHat=jClu+3;
	  // If this is an anti fermion -> use m2rb for g->qq splitting
	  if (idClu[jrHat] < 0 ) {
	    QE2Hat = min(QE2Hat,m2(pClu[jrHat],pClu[jbHat]));
	    mD2Hat = min(mD2Hat,m2(pClu[jrHat],pClu[jbHat]));
	  }
	  // Else if this is a fermion, use m2ar for g->qq splitting
	  else if (idClu[jrHat] < 20) {
	    QE2Hat = min(QE2Hat,m2(pClu[jaHat],pClu[jrHat]));
	    mD2Hat = min(mD2Hat,m2(pClu[jaHat],pClu[jrHat]));
	  }
	  // Else use individual QE, pT, or mD measure for g emission
	  else {
	    double QE2 = RPtr->Q2E(pClu[jaHat],pClu[jrHat],pClu[jbHat]);
	    double mD2 = RPtr->Q2E(pClu[jaHat],pClu[jrHat],pClu[jbHat],2);
	    QE2Hat = min(QE2Hat,QE2);
	    mD2Hat = min(mD2Hat,mD2);
	  }
	}	  
	// Standard Pimp factor
	Pimp         = QE2Hat/(QE2Hat+QE2New);
	// Pimp factor with quartic slope (used for uncertainty bands)
	PimpStronger = pow2(QE2Hat)/(pow2(QE2Hat)+pow2(QE2New));
	// Pimp factor for smooth mD ordering (used for uncertainty bands)
	PimpMD       = mD2Hat/(mD2Hat+mD2New);
	if (verbose >= 8) 
	  printOut("VM::getPME","Pimp = "+num2str(Pimp));
      }
      
      // Compute ARIADNE factor (for g->qqbar splittings)
      double Pari = 1.0;
      if (!sectorShower && (antIndx == iQGsplit || antIndx == iGGsplit)) {
	double sN;
	if (!swapAB) {
	  // Only apply if neighbor is in final state
	  if (ir+1 < int(pClu.size()) ) { 
	    sN = m2(pClu[ir],pClu[ir+1]);
	    // Only apply if neighbor is a quark
	    if (idClu[ir+1] != 21) 
	      Pari = 2.0*sN/(sP + sN);
	  }
	} else { 
	  // Only apply if neighbor is in final state
	  if (ir-2 >= nIn) {
	    sN = m2(pClu[ir-2],pClu[ir-1]);	
	    // Only apply if neighbor is a quark
	    if (idClu[ir-2] != 21)  
	      Pari = 2.0*sN/(sP + sN);
	  }
	}
	
	// Verbose output
	if (verbose >= 8) 
	  printOut("VM::getPME","Pari = "+num2str(Pari));
	
      }
      
      // Get color-ordered M2 (= LC + ...) for this (n-1)-parton state
      //Here, need to sum over possible spin clustering histories
      //only sum over multiple histories if all particles are polarized
      int polClust = 1;
      if (isPol) {
        polClust = (int)polClu.size();
        if (verbose >= 8) 
          cout<<" Looping over "<<polClust<<" internal helicities"<<endl;
      }
      for (int ipol = 0; ipol < polClust; ++ipol) {
        double Mp2;
	if (!isPol) Mp2 = mgInterfacePtr->ME2(nIn, idClu, pClu);
	else Mp2 = mgInterfacePtr->ME2(nIn, idClu, pClu, polClu[ipol]);
	if (Mp2 <= 0.0) {
	  string idString=num2str(idClu[0])+" -> ";
	  string polString=num2str(polClu[ipol][0])+" -> ";
	  for (unsigned int i=1;i<idClu.size();i++) 
	    idString+=num2str(idClu[i]);
	  for (unsigned int i=1;i<polClu[ipol].size();i++) 
	    polString+=num2str(polClu[ipol][i]);
	  printOut("VM::getPME","Unphysical momenta or unknown ME "
		   "for "+idString);
	  printOut("VM::getPME","                                 "
		   "hel "+polString);
	  // Print momentum configuration
	  if (verbose >= 3) {
	    for (unsigned int i=0;i<pClu.size();i++) {
	      cout<< "    i = "<<i<<" px = "<<num2str(pClu[i].px())<<" py = "
		  <<num2str(pClu[i].py())<<" pz = "<<num2str(pClu[i].pz())<<
		" e = "<<num2str(pClu[i].e())<<" m = "<<pClu[i].mCalc()<<endl;
	    }
	  }
        }
      
        // TMP!
        if (verbose >= 9) 
          printOut("VM::getPME","Mp2 = "+num2str(Mp2));
	
        // Check if return was -1 ?
        if (abs(Mp2+1.0) < TINY) return false; 
      
        // Loop over different antenna sets (for uncertainty evaluation)
        // Note: we do not make modifications for strong ordering here, since
        // we assume unordered generation up to the matched orders, and hence
        // the matching corrections should always include unordering. 
        // We also do not make variations in alphaS here, since all powers
        // of the coupling are factored out of the matrix element correction.

	// Set polarizations	
	int polA(9),polB(9),poli(9),polj(9),polk(9);
	if (isPol) {
	  polA = (!swapAB) ? polClu[ipol][ir-1] : polClu[ipol][ir];
	  polB = (!swapAB) ? polClu[ipol][ir] : polClu[ipol][ir-1];
	  poli = (!swapAB) ? pol[ir-1] : pol[ir+1] ;
	  polj = pol[ir];
	  polk = (!swapAB) ? pol[ir+1] : pol[ir-1];
	}

        // Compute antenna function values for USR antennae and optionally
        // for uncertainty variations. Use DEF as basis for variations.
        vector<double> antVal;
        antVal.push_back(getAnt(antIndx,sectorShower,0)->
                          antennaFunction(yar,yrb,mMua,mMur,mMub,
                                          polA,polB,poli,polj,polk)/sP); 
        if (uncertaintyBands) {
          antVal.push_back(getAnt(antIndx,sectorShower,1)->
                            antennaFunction(yar,yrb,mMua,mMur,mMub,
                                            polA,polB,poli,polj,polk)/sP); 
          antVal.push_back(getAnt(antIndx,sectorShower,2)->
                            antennaFunction(yar,yrb,mMua,mMur,mMub,
                                            polA,polB,poli,polj,polk)/sP); 
          antVal.push_back(getAnt(antIndx,sectorShower,3)->
                            antennaFunction(yar,yrb,mMua,mMur,mMub,
                                            polA,polB,poli,polj,polk)/sP); 
        }

        // Loop over uncertainty variations (0 = USR, 1 = DEF, ...)
        for (int iVar=0; iVar<nVariations; iVar++) {	

          // Get saved value of physical antenna function
          double antPhys = antVal[0];
          if (iVar >= 1) antPhys = antVal[1];
          if (iVar == 4) antPhys = antVal[2];
          if (iVar == 5) antPhys = antVal[3];

          // Color/charge factor
          double chargeFactor = 
            getAnt(antIndx,sectorShower,0)->getChargeFactor();
          if (iVar >= 1) chargeFactor = 
                           getAnt(antIndx,sectorShower,1)->getChargeFactor();
          if (iVar == 10 && id[ir] == 21) chargeFactor = CA;
          if (iVar == 11 && id[ir] == 21) chargeFactor = CF;
	  
	  // Uncertainty variations: global shower variation of Pimp factor
          
	  if (!sectorShower) {
	    if (iVar == 8) antPhys *= PimpStronger/Pimp;
	    if (iVar == 9) antPhys *= PimpMD/Pimp;
	  }

	  // Multiply by charge, Pimp (global), and Pari factors
	  antPhys *= Pimp * Pari * chargeFactor;  

          // verbose output
          if ( (verbose >= 8 && iVar == 0) || verbose >= 9) {
            cout<<"VM: antPhys("<<polA<<polB<<">"<<poli<<polj<<polk<<") = "
                <<antPhys<<", isSwapped = "<<bool2str(swapAB)
                <<" m(ar,rb) = "<<sqrt(yar*sP)<<", "<<sqrt(yrb*sP);
            if (iVar >= 1) cout <<" (iVar = "<<iVar<<")"<<endl;
            cout<<endl;
          }

	  // Add to Approximated Mn2 for this set
	  Mn2Appr[iVar] += Mp2 * antPhys;
        }

      //End loop over spin clusterings
      }
    
    }
    
    // Store ME reweight factors
    for (int iVar=0; iVar<nVariations; iVar++) {
      if (iVar >= 1 && ! uncertaintyBands) continue;
      PME.push_back(Mn2/Mn2Appr[iVar]);
    }

    // Diagnostics if shower VERY bad approximation
    if (PME[0] > 1000.0 || (verbose >= 3 && PME[0] > 100.0 )) {
      printOut("VM::getPME","Warning: large correction at nLegs ="
	       +num2str(int(p.size()-1),2)
	       +" unregulated PME[0] = "+num2str(PME[0])
	       +" ME"+num2str(int(p.size()-1),1)+" = "+num2str(Mn2)
	       +" Sum(PS*ME"+num2str(int(p.size()-2),1)+") = "+num2str(Mn2Appr[0]));
    }

    // Store Old ME (future possibility for speedup)
    // oldME2 = Mn2;
   
    // Done
    done = true;
    
    // For the time being, no non-MadGraph matching implemented
    return done;

  }

  //*******

  // getV1():

  // Get a single V1 coefficient, for:
  //    V1[0]        = matching factor renormalization scale s
  //    V1[1]        =
  //    V1[2]        = 
  //    V1[3]        =
  
  bool Matching::getV1(int nIn, vector<Particle*> particlePtrs, 
                       vector<double>& V1) {

    // Default value = 1
    bool V1bool = false;
    
    // Sanity checks
    if (nIn != 1) return V1bool;

    // Require at least 3 particles
    if (particlePtrs.size() <= 2) return V1bool;

    // Settings extracted from Vincia
    int evVarEmit = evolutionType;
    int ordeMode  = orderingMode;

    // Data about final state
    int idRes     = particlePtrs[0]->id();
    int nFS       = particlePtrs.size() - nIn;
    // TMP!!! Insert real running alphaS value later.
    double alphaS = 0.12;

    // Global Invariants
    double mSys   = particlePtrs[0]->m();
    double m2sys  = pow2(mSys);
    //double s      = pow2(mSys);
    //double nf     = getNf(mSys);
    double nf = 5;

    // Z decays
    if (idRes == 23) {

      // Z -> 2
      if (nFS == 2){ 
        V1[0]=alphaS/M_PI;
        V1bool = true;
      }

      // Z -> 3
      else if (nFS == 3) {

        // Invariants : Q_I G_J Qbar_K
        double sij = m2(particlePtrs[1]->p() + particlePtrs[2]->p());
        double sjk = m2(particlePtrs[2]->p() + particlePtrs[3]->p());
        //double sik = m2(particlePtrs[1]->p() + particlePtrs[3]->p());
        double yij = sij/m2sys;
        double yjk = sjk/m2sys;
        
	// NLO matching

	//Fixed Order contribution
	double Born,RfunFull1,RfunFull2,RfunFull3;

	Born = yij/yjk + (2*(1 - yij - yjk))/(yij*yjk) + yjk/yij;
	RfunFull1 = (nf*(log(yij) + log(yjk)))/6.;
	RfunFull2 = CA*(-4 - pow2(M_PI)/6. - (5*log(yij))/3. + 
                    log(1 - yij)*log(yij) - (5*log(yjk))/3. 
                    - log(yij)*log(yjk) + log(1 - yjk)*log(yjk) + 
			Li2(yij) + Li2(yjk));
	RfunFull3 =(CA*(1 + (1 - yij)/(2.*yij) - yij/(2.*(1 - yjk)) + (1 - yjk)/(2.*             yjk) - yjk/(2.*(1 - yij)) + (2 + (2*(yij - yjk))/(1 - yij) - 
                    (yij*yjk)/(2.*pow2(1 - yij)))*log(yij) + 
                    (2 - (yij*yjk)/(2.*pow2(1 - yjk)) + 
                     (2*(-yij + yjk))/(1 - yjk))*log(yjk)))/Born;

        if( verbose == 9 ){
          printOut("VM::getV1","Loop calculation = "
                                    +num2str(RfunFull1+RfunFull2+RfunFull3));
	}


	// Renormalization terms (PS TODO: need to generalize CMW factor to 
        // arbitrary numbers of flavors)
	double kMu2 = 1.;
        if (showerAlphaScmw) kMu2 /= pow2(1.569);
	double SLmuRpT;
        //double SLmuRGeo, SLmuRmD;
        
        SLmuRpT = (11/2. - nf/3.)*log(1./(kMu2*yij*yjk))*alphaS/2./M_PI;
        //SLmuRGeo = (11/2. - nf/3.)*log(sqrt(1/(sij*sjk))/kMu2)*alphaS/2./M_PI;
        if( sij < sjk ){
	  //SLmuRmD = (11/2. - nf/3.)*log(1/(2.*kMu2*sij))*alphaS/2./M_PI;
        }
	else if( sjk <= sij ){
	  //SLmuRmD = (11/2. - nf/3.)*log(1/(2.*kMu2*sjk))*alphaS/2./M_PI;
	}

	// Ariadne log
	double AriadneLog;
	AriadneLog = -nf*(yij-yjk)/(yij+yjk)*log(yij/yjk)/6.;

        if( verbose == 9 ){
          printOut("VM::getV1()","AriadneLog = "
                                                       +num2str(AriadneLog));
	}

	//VinciaSubtractionTerms Sud23StpT;
        double N;
        // N forefactor of pimp factor
        // make dynamic --> ask peter
        N = 4;
	double Sud23StpT, Sud34GESmpT, Sud34GSSmpT;
        subtractionTerms.init();
        Sud23StpT = subtractionTerms.IqqEmit(yij,yjk,evVarEmit);
        Sud34GESmpT = subtractionTerms.IqgEmit(yij,yjk,evVarEmit,ordeMode);
        Sud34GSSmpT = subtractionTerms.IqgSplit(yij,yjk,evVarEmit,ordeMode,N);

        if( verbose == 9 ){
          printOut("VM::getV1()","2->3 Shower integrals = "
                                                       +num2str(Sud23StpT));
          printOut("VM::getV1()","3->4 GE Shower integrals = "
                                                       +num2str(Sud34GESmpT));
          printOut("VM::getV1()","3->4 GS Shower integrals = "
                                                       +num2str(Sud34GSSmpT));
	}

        // born level matching
        double V0 = 2.;

        // finite terms arising from integrating antenna over full phase space
        double intd30, inte30;
        intd30 = 34/3.*CA;
        inte30 = -nf;        

        // finite terms arising from the tree-level matching corrections
        // (the deltaA terms, representing the deviation from standard GGG)
        // TODO: should be delegated to a separate function
        double sumEmit     = 0.;
        double sumSplit    = 0.;
        double comFacEmit  = CA/4.;
        double comFacSplit = nf/4.;

        // Store original momentum configuration as 4-vectors
        vector<Vec4> pOrg;
        for (int iOrg=1; iOrg<=3; ++iOrg) 
          pOrg.push_back(particlePtrs[iOrg]->p());
        double antOrg = functionsPtr->A30(pOrg,0,1,2);

        // Do MC loop
        int nPointsMC = 100;
        for (int i=0; i<nPointsMC; ++i) {
          // The n-parton phase-space point already uses yij and yjk
          // Use xij and xjk to integrate over (n+1)-parton phase space
          vector<double> x;
          x.push_back(rndmPtr->flat());
          x.push_back(rndmPtr->flat());
          if (x[0] + x[1] > 1.0) {
            x[0] = 1.-x[0];
            x[1] = 1.-x[1];
          }
          double phi = 2. * M_PI * rndmPtr->flat();
 
          // Sum over radiating antennae.
          vector<Vec4> nPlusOne;
          for (int iA = 0; iA < int(pOrg.size()) -1 ; ++iA) {

            // Set up radiating system: [iA,iA+1]
            vector<Vec4> pTwo;
            pTwo.push_back(pOrg[iA]);
            pTwo.push_back(pOrg[iA+1]);
                        
            //-------------------------------
            // QG Emit (and GQbar, with swap)
            int iAnt = iQGemit;
            bool swap = false;
            if (iA == 1) swap = true;

            // Compute antennae. Conjugate if needed.
            Antenna* antPtr = antennaSets[0]->getAnt(iAnt);
            double aVincia  = antPtr->antennaFunction(x[swap],x[1-swap]);
            double aStd     = subtractionTerms.getA(iAnt,x[swap],x[1-swap]);

            // Evaluate Pimp factor
            double ordE = 1.0;
            // Assume smooth pT-ordering for now
            if (iA == 0) {
              ordE = yjk/(yjk + x[0]*x[1]);
            } else {
              ordE = yij/(yij + x[0]*x[1]);
            }

            // Construct (n+1)-parton phase-space point
            int kineMapEmit = 1;          
            kineMapEmit = antPtr->getKineMapType();
            vector<Vec4> pThree;
            vinCluPtr->map2to3light(pThree,pTwo,kineMapEmit,x[0],x[1],phi);
            // Store new particles in vector that can be passed the LO ME
            nPlusOne.resize(0);
            for (int iCopy=0; iCopy<int(pOrg.size()); ++iCopy) {
              if (iCopy != iA && iCopy != iA+1) nPlusOne.push_back(pOrg[iCopy]);
              else if (iCopy == iA) {
                // Insert branched momenta in place of unbranched ones
                nPlusOne.push_back(pThree[0]);
                nPlusOne.push_back(pThree[1]);
                nPlusOne.push_back(pThree[2]);
              }
            } 
            // Numerator: Leading-Color (n+1)-parton matrix element
            double antMELC = functionsPtr->A40(nPlusOne,0,1,2,3);

            // Denominator: antenna approximation (start with this clustering)
            double antSumVincia = ordE * aVincia * antOrg;
            // Global showers: also sum over alternative clusterings
            for (int ir = 1; ir < int(nPlusOne.size())-1; ++ir) {
              if (ir == iA+1) continue;
              vector<Vec4> pClu(pOrg);                
              vinCluPtr->map3to2(kineMapEmit, nPlusOne[ir-1],
                                   nPlusOne[ir], nPlusOne[ir+1],
                                   pClu[ir-1], pClu[ir]);
              // Compute 2->3 antenna for the reclustered point
              double antClu = functionsPtr->A30(pClu,0,1,2);
              // Compute 3->4 antenna for this clustering
              double sNow = m2(pClu[ir-1],pClu[ir]);
              vector<double> xNow;
              xNow.push_back(m2(nPlusOne[ir-1],nPlusOne[ir]) / sNow);
              xNow.push_back(m2(nPlusOne[ir],nPlusOne[ir+1]) / sNow);
              int iAntNow = iQGemit;
              bool swapNow = false;
              if (ir == 2) swapNow = true;
              // Compute antennae. Conjugate if needed.
              Antenna* antPtrNow = antennaSets[0]->getAnt(iAntNow);
              double aVinciaNow  = 
                antPtrNow->antennaFunction(xNow[swapNow],xNow[1-swapNow]);
              // Compute ordering factor (assume smooth pT-ordering for now)
              double q2now = xNow[0] * xNow[1] * sNow;
              double q2clu = m2(pClu[0],pClu[1])*m2(pClu[1],pClu[2]) / 
                m2(pClu[0] + pClu[1] + pClu[2]);
              double ordClu = q2clu/(q2clu + q2now);
              antSumVincia += ordClu * aVinciaNow * antClu;
            }
            double Rtree = antMELC/antSumVincia;
            sumEmit  += ordE * (Rtree*aVincia-aStd);
            
            //-------------------------------
            // QG Split (and GQbar with swap)
          /*
            if (hasAntenna[iQGsplit]) {
             // Evaluate Pimp*PAri factor
              double ordS = 1.0;
              // Assume smooth pT-ordering for now
              if (iA == 1) {
                ordS = 2. * pTnormalization * pow2(yjk) 
                  / (pTnormalization*yjk + xjk) / (yjk + yij);
              } else {
                ordS = 2. * pTnormalization * pow2(yij) 
                  / (pTnormalization*yij + xjk) / (yjk + yij);
              }
              // Recycle phase-space point if same kineMapType
              int kineMapSplit = antPtr[iQGsplit]->getKineMapType();
              if (kineMapSplit != kineMapEmit) {
                vector<Vec4> pThree;
                vinCluPtr->map2to3light(pThree,pTwo,kineMapSplit,xij,xjk,phi);
                // Store new particles in vector that can be passed the LO ME
                nPlusOne.resize(0);
                for (int iCopy=1; iCopy<int(particlePtrs.size()); ++iCopy) {
                  if (iCopy != iA && iCopy != iA+1) 
                    nPlusOne.push_back(particlePtrs[iCopy]->p());
                  else if (iCopy == iA) {
                    // Insert branched momenta in place of unbranched ones
                    nPlusOne.push_back(pThree[0]);
                    nPlusOne.push_back(pThree[1]);
                    nPlusOne.push_back(pThree[2]);
                  }
                } 
              }
              // Compute (n+1)-parton matrix element
              double antMELC = functionsPtr->B40(nPlusOne,0,1,2,3);
              // Compute antenna approximation
              double antSumVincia = antMELC;
              double Rtree = antMELC/antSumVincia;
              sumSplit += ordS * (Rtree * aVincia[iQGsplit] - aStd[iQGsplit]);
              
              } // end if hasSplit
          */
          } // end sum over antennae
        } // end Monte Carlo integration loop
        double deltaA = (comFacEmit*sumEmit + comFacSplit*sumSplit) / nPointsMC;

        // building the expanded sudakov
        double shower;
        shower = Sud23StpT - Sud34GESmpT - Sud34GSSmpT + intd30 + inte30 - V0
          + deltaA;

        //Matching term (without renormalization)
        double alphaSfac = alphaS/2./M_PI;
	double matchNLOnomuR = alphaSfac * ( shower + RfunFull1 + RfunFull2 
                                             + RfunFull3 + AriadneLog);

	//V1[0] = matchNLO_nomuR;
	V1.resize(1);
	V1[0] = matchNLOnomuR - SLmuRpT;
	if (uncertaintyBands) {
 	//V1[2] = matchNLO_nomuR-SLmuRGeo;
	//V1[3] = matchNLO_nomuR-SLmuRmD;
	}
	V1bool = true;

        //cout<<" V1 = "<<V1[0]<<" S "<<shower<<" of which dA = "<<deltaA<<" R "<<RfunFull1<<" "<<RfunFull2<<" "<<RfunFull3<<" "<<AriadneLog<<" mu "<<-alphaSfac*SLmuRpT<<" "<<endl;

        // TMP: print terms and size of deltaA
        if ( verbose == 9 ) {
          cout<<" terms = "<<shower<<" "<<RfunFull1<<" "<<RfunFull2<<" "<<RfunFull3<<" "<<AriadneLog<<endl;
          cout<<"NLO corr = "<<num2str(V1[0],6)<<" ; size of deltaA = "<<num2str(deltaA*alphaS/2./M_PI,6) << endl;

          printOut("VM::getV1","Shower expansion = "+num2str(shower));
          printOut("VM::getV1","   matchNLOnomuR = "+num2str(matchNLOnomuR));
          printOut("VM::getV1","Matching factor with muR=pT = "
                   +num2str(V1[0]));
	}

      }
      
    }

    return V1bool;

  }
  

  //*******

  // get2to4():
  // Return 2->4 reweighting factors 
  //   id  : id codes for the 4 partons
  //   p   : 4-momenta for the 4 partons
  // Return reweighting factors in PME
  //   P2to4[0] : reweighting factor with USR antennae
  //   P2to4[n] : reweighting factor with variations, see getPME()
  // = |A_4|^2 / (Sum_i A^phys_i A^phys_i_n-1)

  bool Matching::getP2to4(vector<int> id, vector<Vec4> p, 
			vector<double>& P2to4) {

    // Resize P2to4.
    P2to4.resize(0);
    bool done = false;

    // Check if at least 4 partons in final state
    if (id.size() <= 5) return false;

    // Initialize (n-1)->n vector of approximations
    vector<double> Mn2Appr(0);
    for (int iVar=0; iVar<nVariations; iVar++) {
      Mn2Appr.push_back(0.);
    }
    
    // SYSTEMATIC 2->4 MATCHING
    // Requires that collinear singularities correctly
    // partitioned in 2->4 functions

    // Loop over possible emitter antennae
    for (int iClu=0; iClu<int(id.size())-3; iClu++) {
      
      // Clustering:
      //cout<<" P2to4 Clustering "<<iClu<<"/"<<int(int(id.size())-4)<<endl;
      
      // Compute new ordering scale
      int ir  = 2+iClu;
      int ida = id[ir-1];
      int idr = id[ir];
      int idb = id[ir+1];
      
      //cout <<"  arb = "<<ir-1<<" "<<ir<<" "<<ir+1<<" QNew = "<<sqrt(Q2New)<<endl;
      
      // B40A: (q qbar) -> (q g) qbar -> q q'bar q' qbar
      // B40B: (q qbar) -> q (g qbar) -> q q'bar q' qbar
      
      // E40aA: (q g) -> (q qbar' q') -> q g qbar' q'
      // (match after gluon emission inside qqbar antenna)
      // E40aB: (q g) -> q (g g) -> q g qbar' q'
      // (match after gluon splitting inside gg antenna, outside)
      
      // E40bA: (q g) -> (q g) g -> q q'bar q' g
      // (match after gluon splitting inside qg antenna
      // E40bB: (q g) -> q (g g) -> q q'bar q' g
      // (match after gluon splitting inside gg antenna, inside)
      
      // H40A: (g g) -> qbar (q g) -> qbar q qbar' q'
      // (match after gluon splitting inside qg antenna
      // H40B: (g g) -> (g qbar) q -> qbar q qbar' q'
      // (match after gluon splitting inside gqbar antenna
      
      // G40A: (g g) -> (g g) g -> g qbar q g
      // (match after gluon splitting inside qg antenna
      // G40B: (g g) -> g (g g) -> g qbar q g
      // (match after gluon splitting inside gqbar antenna
      
      // Reconstruct previous state
      vector<int> idClu(id);
      vector<Vec4> pClu;
      int idaHat  = ida;
      int idbHat  = idb;
      int antIndx = -1;
      bool swapAB = false;
      if (idr == 21) {
	// To recluster gluon, just collapse parent ids
	idClu.erase(idClu.begin()+ir);
	if (ida == 21 && idb == 21) antIndx = iGGemit;
	else if (idb == 21) antIndx = iQGemit;
	else if (ida == 21) {antIndx = iQGemit; swapAB=true;}
	else antIndx = iQQemit;
      } else if (idr < 0) {
	// To recluster antiquark, collapse with downstream quark 
	idClu[ir+1]=21;
	idClu.erase(idClu.begin()+ir);
	idbHat = 21;
	if (idaHat == 21) antIndx = iGGsplit;
	else antIndx = iQGsplit;
      } else {
	// To recluster quark, collapse with upstream antiquark
	idClu[ir-1]=21;
	idClu.erase(idClu.begin()+ir);
	idaHat = 21;
	if (idbHat == 21) {antIndx = iGGsplit; swapAB=true;}
	else {antIndx = iQGsplit; swapAB=true;}
      }
      if (verbose >= 9) {
	cout << "P2to4 Clustering "<<iClu<<" : ";
	for (int i=0;i<int(idClu.size());i++) cout << " " << idClu[i];
	if (antIndx == iQGemit) cout << (swapAB ? " gqbarEmit" : " qgEmit");
	cout<<" Ant = "<<antennaSets[0]->getAnt(antIndx)->humanName();
	cout<<endl;
      }
      // Reconstruct clustered momenta
      int iMap = antennaSets[0]->getAnt(antIndx)->getKineMapType();
      vinCluPtr->vc3to2(iMap,p,ir-1,ir,ir+1,pClu);
      if (verbose >= 9) {
	for (int i=0;i<int(idClu.size());i++) {
	  cout << "   p" << i<<" "<<pClu[i].px()<<" "<<pClu[i].py()
	       <<" "<<pClu[i].pz()<<" "<<pClu[i].e()<<endl;
	}
      }
      
      // Invariant mass of parent antenna and daughter invariants
      double sP  = m2(pClu[ir-1],pClu[ir]);
      double yar = m2(p[ir-1],p[ir])/sP;
      double yrb = m2(p[ir],p[ir+1])/sP;
      if (swapAB) {
	double ytmp = yar;
	yar = yrb;
	yrb = ytmp;
      }
      
      // Compute Pimp factor (only used for global showers) 
      //  : orderingMode <= 1 : Pimp = 1
      //  : orderingMode == 2 : Q2E/(Q2Ehat+Q2E) 
      double Pimp    = 1.0;
      // Additional Pimp factor variations used for uncertainty estimates
      // (Not active in 2 to 4 yet)
      //double PimpStronger = 1.0;
      //double PimpMD  = 1.0;

      // Pimp factors only used in global showers
      if ( !sectorShower && orderingMode >= 2 ) { 

	// QE, pT, and mD Values for current clustering	
	double QE2New;
	// If this is an anti fermion -> always use m2rb for g->qqbar
	if (id[ir] < 0) {
	    QE2New = m2(p[ir],p[ir+1]);
	    //mD2New = QE2New;
	}
	// Else if this is a fermion -> always use m2ar for g->qqbar
	if (id[ir] < 0) {
	    QE2New = m2(p[ir-1],p[ir]);
	    //mD2New = QE2New;
	}
	// Else use QE, pT, mD definitions for g emission
	else {
	  QE2New = RPtr->yE(yar,yrb)*sP;
	  //mD2New = RPtr->yE(yar,yrb,2)*sP;
	}

	// Initial values for Hat variables characterizing the reclustered
	// parton state
	double QE2Hat=m2(pClu[1],pClu[2],pClu[3]);
	//double mD2Hat=mD2Hat;
	// Loop over possible clusterings, find smallest Q for each measure
	for (int jClu=0; jClu<int(pClu.size())-3; jClu++) {
	  int jaHat=jClu+1;
	  int jrHat=jClu+2;
	  int jbHat=jClu+3;
	  // If this is an anti fermion -> use m2rb for g->qq splitting
	  if (idClu[jrHat] < 0 ) {
	    QE2Hat = min(QE2Hat,m2(pClu[jrHat],pClu[jbHat]));
	    //mD2Hat = min(mD2Hat,m2(pClu[jrHat],pClu[jbHat]));
	  }
	  // Else if this is a fermion, use m2ar for g->qq splitting
	  else if (idClu[jrHat] < 20) {
	    QE2Hat = min(QE2Hat,m2(pClu[jaHat],pClu[jrHat]));
	    //mD2Hat = min(mD2Hat,m2(pClu[jaHat],pClu[jrHat]));
	  }
	  // Else use individual QE, pT, or mD measure for g emission
	  else {
	    double QE2 = RPtr->Q2E(pClu[jaHat],pClu[jrHat],pClu[jbHat]);
	    //double mD2 = RPtr->Q2E(pClu[jaHat],pClu[jrHat],pClu[jbHat],2);
	    QE2Hat = min(QE2Hat,QE2);
	    //mD2Hat = min(mD2Hat,mD2);
	  }
	}	  
	// Standard Pimp factor 
	Pimp  = QE2Hat/(QE2Hat+QE2New);
	// Pimp factor with quartic slope (used for uncertainty bands)
	//PimpStronger = pow2(QE2Hat)/(pow2(QE2Hat)+pow2(QE2New));
	// Pimp factor with MD2 for gluon emissions (used for uncertainty bands)
	//PimpMD  = mD2Hat/(mD2Hat+mD2New);
	if (verbose >= 8) 
	  printOut("VM::getP2to4","Pimp = "+num2str(Pimp));
      }
      
      // Compute ARIADNE factor (for g->qqbar splittings)
      double Pari = 1.0;
      if (!sectorShower && (antIndx == iQGsplit || antIndx == iGGsplit)) {
	double sN;
	if (!swapAB) {
	  // Only apply if neighbor is in final state
	  //	  if (ir+1 < int(pClu.size()) ) { 
	  sN = m2(pClu[ir],pClu[ir+1]);
	  // Only apply if neighbor is a quark
	  if (idClu[ir+1] != 21) 
	    Pari = 2.0*sN/(sP + sN);
	  //	  }
	} else { 
	  // Only apply if neighbor is in final state
	  //	  if (ir-2 >= nIn) {
	  sN = m2(pClu[ir-2],pClu[ir-1]);	
	  // Only apply if neighbor is a quark
	  if (idClu[ir-2] != 21)  
	    Pari = 2.0*sN/(sP + sN);
	  //	  }
	}
	
	// Verbose output
	if (verbose >= 8) 
	  printOut("Functions::getP2to4","Pari = "+num2str(Pari));
	
      }
      
      // Get color-ordered M2 (= LC + ...) for this (n-1)-parton state
      double Mp2 = 0.0 ; // mgInterfacePtr->ME2(nIn, idClu, pClu);
      
      // TMP!
      if (verbose >= 9) {
	cout << " P2to4 Mp2 = "<<Mp2<<endl;
	
	// Test
	double sC01  = m2(pClu[1],pClu[2]);
	double sC12  = m2(pClu[2],pClu[3]);
	double sC    = m2(pClu[1],pClu[2],pClu[3]);
	double A3chk = CF*antennaSets[0]->getAnt(iQQemit)->
        antennaFunction(sC01/sC,sC12/sC)/sC;
	double A3chk2 = antennaSets[0]->getAnt(iQQemit)->
        getChargeFactor()
	  *antennaSets[0]->getAnt(iQQemit)->antennaFunction(sC01/sC,sC12/sC)/sC;
	cout << " P2to4 A30 = "<<scientific<<A3chk<<" s = "<<" A30chk = "
	     << A3chk2<<endl;
      }
      
      // Check if return was -1 ?
      if (abs(Mp2+1.0) < TINY) return false; 
      
      // Loop over different antenna sets (for uncertainty evaluation)
      // Use DEF set as basis for uncertainty variations
      vector<double> antVal;
      antVal.push_back(getAnt(antIndx,sectorShower,0)->
                        antennaFunction(yar,yrb)/sP); 
      if (uncertaintyBands) {
        antVal.push_back(getAnt(antIndx,sectorShower,1)->
                          antennaFunction(yar,yrb)/sP); 
        antVal.push_back(getAnt(antIndx,sectorShower,2)->
                          antennaFunction(yar,yrb)/sP); 
        antVal.push_back(getAnt(antIndx,sectorShower,3)->
                          antennaFunction(yar,yrb)/sP); 
      }
      for (int iVar=0; iVar<nVariations; iVar++) {	
	// Compute physical antenna for this set for this (n-1)-parton state
	double antPhys = antVal[0];
        if (iVar >= 1) antPhys = antVal[1];
        if (iVar == 4) antPhys = antVal[2];
        if (iVar == 5) antPhys = antVal[3];
	double chargeFactor = getAnt(antIndx,sectorShower,0)->getChargeFactor();
        if (iVar >= 1) 
          chargeFactor = getAnt(antIndx,sectorShower,1)->getChargeFactor();
	if (iVar == 10 && id[ir] == 21) chargeFactor = CA;
	if (iVar == 11 && id[ir] == 21) chargeFactor = CF;
	antPhys *= chargeFactor;
	antPhys *= Pimp * Pari;
	// Add to Approximated Mn2 for this set
	Mn2Appr[iVar] += Mp2 * antPhys;	
      }
      
    }
    
    // Store ME reweight factors
    for (int iSet=0; iSet<int(antennaSets.size()); iSet++) {
      if (iSet >= 1 && ! uncertaintyBands) continue;
      //P2to4.push_back(Mn2/Mn2Appr[iSet]);
    }
    
    // Store Old ME
    //oldME2 = Mn2;
    
    
    // Done
    done = true;
    
    // For the time being, no non-MadGraph matching implemented
    return done;

  }

}
