// VRdisplay.cc is a part of the VINCIA plug-in to the PYTHIA 8 generator.
// Copyright (C) 2013 P. Skands.
// VINCIA is licenced under the GNU GPL version 2.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// Function definitions (not found in the header) for the VRdisplay class 

// Include the VinciaRootDisplay header
#include "VRdisplay.h"

namespace Vincia {
  
  // Init pointers and set settings
  void VRdisplay::init(VRsettings* settingsPtrIn) {
    settingsPtr   = settingsPtrIn;
    logY          = settingsPtr->logY;
    normalize     = settingsPtr->normalize;
    normFac       = settingsPtr->normFac;
    optStat       = settingsPtr->optStat; 
    verbose       = settingsPtr->verbose;
    sigmaStat     = settingsPtr->sigmaStat;
    nPad          = 1;      
    setShowRatio(settingsPtr->showRatioPad);
    setShowRelUnc(settingsPtr->showRelUncPads);
    setMCnames(settingsPtr->MCnames);
    setLegend(settingsPtr->legend);
    setSignature(settingsPtr->signature);
    setLeftLegend(false);
  }

  // Add pad (for ratio to data or decomposition of vector of histos)
  void VRdisplay::addPad() {
    nPad++;
    // When adding many pads, increase label font sizes
    for (int i=0; i<nPad-1; ++i) {
      double labSizeX = 1.07 * axesPtr->GetXaxis()->GetLabelSize();
      double labSizeY = 1.07 * axesPtr->GetYaxis()->GetLabelSize();
      axesPtr->GetXaxis()->SetLabelSize(labSizeX);
      axesPtr->GetYaxis()->SetLabelSize(labSizeY);      
      double titSizeX = 1.09 * axesPtr->GetXaxis()->GetTitleSize();
      double titSizeY = 1.09 * axesPtr->GetYaxis()->GetTitleSize();
      axesPtr->GetXaxis()->SetTitleSize(titSizeX);
      axesPtr->GetYaxis()->SetTitleSize(titSizeY);      
      double titOffsetY = 0.94*axesPtr->GetYaxis()->GetTitleOffset();
      axesPtr->GetYaxis()->SetTitleOffset(titOffsetY);      
      // Ratio pad
      double labNormSizeX = 1.05 * axesNormPtr->GetXaxis()->GetLabelSize();
      double labNormSizeY = 1.05 * axesNormPtr->GetYaxis()->GetLabelSize();
      axesNormPtr->GetXaxis()->SetLabelSize(labNormSizeX);
      axesNormPtr->GetYaxis()->SetLabelSize(labNormSizeY);      
      double titNormSizeX = 1.18*axesNormPtr->GetXaxis()->GetTitleSize();
      double titNormSizeY = 1.10*axesNormPtr->GetYaxis()->GetTitleSize();
      axesNormPtr->GetXaxis()->SetTitleSize(titNormSizeX);
      axesNormPtr->GetYaxis()->SetTitleSize(titNormSizeY);      
      double titNormOffsetY = 0.9*axesNormPtr->GetYaxis()->GetTitleOffset();
      axesNormPtr->GetYaxis()->SetTitleOffset(titNormOffsetY);      
    }
    // RelUnc pad(s)
    if (nPad >= 3) {
      double labVarSizeX = 1.75 * axesVarPtr->GetXaxis()->GetLabelSize();
      double labVarSizeY = 1.75 * axesVarPtr->GetYaxis()->GetLabelSize();
      axesVarPtr->GetXaxis()->SetLabelSize(labVarSizeX);
      axesVarPtr->GetYaxis()->SetLabelSize(labVarSizeY);      
      double titVarSizeX = max(1.75 * axesVarPtr->GetXaxis()->GetTitleSize(),
                               double(axesVarPtr->GetXaxis()->GetLabelSize()));
      double titVarSizeY = 1.85 * axesVarPtr->GetYaxis()->GetTitleSize();
      double titVarOffsetY = 0.6 * axesVarPtr->GetYaxis()->GetTitleOffset();
      axesVarPtr->GetXaxis()->SetTitleSize(titVarSizeX);
      axesVarPtr->GetYaxis()->SetTitleSize(titVarSizeY);      
      axesVarPtr->GetYaxis()->SetTitleOffset(titVarOffsetY);
      isInitPad = false;
    }
  }

  // Goto pad
  void VRdisplay::gotoPad(int iPadIn) {      
    displayCanvasPtr->cd();      
    
    // If pads already defined, just cd to the desired one
    if (isInitPad) {
      gPad->cd(min(iPadIn,nPad));
      return;
    }

    // If pads not yet defined, create them 
    gPad->Divide(1,nPad);
    double nDiv = 4. + nPad;
    double topMargin    = 0.5*gPad->GetTopMargin();
    double bottomMargin = gPad->GetBottomMargin();
    if (nPad >= 2) bottomMargin *= 1.75 + 0.5*nPad;
    double leftMargin   = (1.05+nPad/10.)*gPad->GetLeftMargin();
    double rightMargin  = 0.5*gPad->GetRightMargin();      
    for (int iPad=1; iPad<=nPad; ++iPad) {
      displayCanvasPtr->cd();        
      gPad->cd(iPad);
      double yHi = 1.0;
      if (iPad >= 2) yHi = 1.0 - double(2+iPad)/nDiv;
      double yLo = 1.0 - double( (iPad+3)/nDiv );
      if (iPad == nPad) yLo = 0.0;
      // Set pad coordinates
      gPad->SetPad(0.0,yLo,1.0,yHi);
      // Set inner margins
      if (iPad == 1) gPad->SetTopMargin(topMargin);
      else gPad->SetTopMargin(0.0);
      if (iPad == nPad) gPad->SetBottomMargin(bottomMargin);
      else gPad->SetBottomMargin(0.0);
      gPad->SetLeftMargin(leftMargin);
      gPad->SetRightMargin(rightMargin);
      if (!isInitPad) {
        gPad->Draw();
      } else {
        gPad->Modified();
      }
    }
    isInitPad = true;
    gPad->cd(min(iPadIn,nPad));
  }
  
  // Check if axes and/or data has been defined
  void VRdisplay::setAxes(TH1D* initHis) {
    if (hasAxes()) {
      delete axesPtr; 
      delete xCtrPtr;
      delete xUppPtr;
      delete axesNormPtr;
      delete axesVarPtr;
    }
    axesPtr = new TH1D(*initHis);
    axesPtr->Reset("ICES");
    hasAxesSav = true;
    // Axis histogram is pure dummy, has no stats, so switch off stats display
    axesPtr->SetStats(false);
    // Set fonts and axis tick divisions 
    axesPtr->GetXaxis()->SetLabelFont(42);
    axesPtr->GetYaxis()->SetLabelFont(42);
    axesPtr->GetXaxis()->SetNdivisions(505);
    axesPtr->GetYaxis()->SetNdivisions(606);
    axesPtr->GetXaxis()->SetTitleFont(42);
    double labSizeX = 1.2 * axesPtr->GetXaxis()->GetLabelSize();
    double labSizeY = 1.2 * axesPtr->GetYaxis()->GetLabelSize();
    axesPtr->GetXaxis()->SetLabelSize(labSizeX);
    axesPtr->GetYaxis()->SetLabelSize(labSizeY);
    axesPtr->GetYaxis()->SetTitleSize(1.1*axesPtr->GetYaxis()->GetTitleSize());
    axesPtr->GetYaxis()->SetTitleOffset(1.18*axesPtr->GetYaxis()->GetTitleOffset());
    axesPtr->GetYaxis()->SetTitleFont(42);
    // Switch off lame ROOT title
    axesPtr->SetTitle("");

    // Set axis histo for ratio pad
    axesNormPtr = new TH1D(*axesPtr);
    axesNormPtr->SetMinimum(0.5);
    axesNormPtr->SetMaximum(1.5);
    // X axis
    axesNormPtr->GetXaxis()->SetTickLength(0.1);
    axesNormPtr->GetXaxis()->SetLabelSize(1.9*axesPtr->GetXaxis()->GetLabelSize());
    axesNormPtr->GetXaxis()->SetTitleFont(42);
    axesNormPtr->GetXaxis()->SetTitleSize(1.85*axesPtr->GetXaxis()->GetTitleSize());
    axesNormPtr->GetXaxis()->SetTitleOffset(
                                            1.*axesPtr->GetXaxis()->GetTitleOffset());
    // Y axis
    axesNormPtr->GetYaxis()->SetLabelFont(42);
    axesNormPtr->GetYaxis()->SetLabelSize(1.8*axesPtr->GetYaxis()->GetLabelSize());
    axesNormPtr->GetYaxis()->SetLabelOffset(0.01);
    axesNormPtr->GetYaxis()->SetTitleFont(42);
    axesNormPtr->GetYaxis()->SetTitleSize(1.8*axesPtr->GetYaxis()->GetTitleSize());
    axesNormPtr->GetYaxis()->SetTitle("Theory/Data");          
    axesNormPtr->GetYaxis()->SetTitleOffset(0.63*axesPtr->GetYaxis()->GetTitleOffset());
    axesNormPtr->GetYaxis()->CenterTitle();
    axesNormPtr->GetYaxis()->SetTickLength(0.025);
    // Make copy of x axis for overdrawing 
    int nBin = axesNormPtr->GetNbinsX();
    xCtrPtr = new TLine(axesNormPtr->GetBinLowEdge(1),1.0,
                        axesNormPtr->GetBinLowEdge(nBin+1),1.0);
    xUppPtr = new TLine(axesNormPtr->GetBinLowEdge(1),1.499,
                        axesNormPtr->GetBinLowEdge(nBin+1),1.499);
    
    
    // Set axis histo for relUnc pads (uncertainty variations)
    axesVarPtr = new TH1D(*axesPtr);
    axesVarPtr->SetMinimum(-1.25);
    axesVarPtr->SetMaximum(1.25);
    // X axis
    axesVarPtr->GetXaxis()->SetTickLength(2.*axesPtr->GetXaxis()->GetTickLength());
    axesVarPtr->GetXaxis()->SetLabelOffset(2.*axesPtr->GetXaxis()->GetLabelOffset());
    axesVarPtr->GetXaxis()->SetLabelSize(2.*axesPtr->GetXaxis()->GetLabelSize());
    axesVarPtr->GetXaxis()->SetTitleFont(42);
    axesVarPtr->GetXaxis()->SetTitleSize(2.5*axesPtr->GetXaxis()->GetTitleSize());
    axesVarPtr->GetXaxis()->SetTitleOffset(
                                           1.*axesPtr->GetXaxis()->GetTitleOffset());
    // Y axis
    axesVarPtr->GetYaxis()->SetNdivisions(203);
    axesVarPtr->GetYaxis()->SetTickLength(axesPtr->GetYaxis()->GetTickLength());
    axesVarPtr->GetYaxis()->SetLabelFont(42);
    axesVarPtr->GetYaxis()->SetLabelSize(1.8*axesPtr->GetYaxis()->GetLabelSize());
    axesVarPtr->GetYaxis()->SetLabelOffset(1.8*axesPtr->GetYaxis()->GetLabelOffset());
    axesVarPtr->GetYaxis()->SetTitleFont(42);
    axesVarPtr->GetYaxis()->SetTitle("Rel. Unc.");          
    axesVarPtr->GetYaxis()->SetTitleSize(2*axesPtr->GetYaxis()->GetTitleSize());
    axesVarPtr->GetYaxis()->SetTitleOffset(0.5*axesPtr->GetYaxis()->GetTitleOffset());
    axesVarPtr->GetYaxis()->CenterTitle();
  }
  
  //========================================================================
  // VRdisplay: FILE I/O
  
  // Set file to be treated as experimental data,
  // with i1 : i12 specifying data file format, columnwise (see VINCIA docs)
  bool VRdisplay::setDataFile(string dataFileName, int i1, int i2, int i3, 
                              int i4, int i5, int i6, int i7, int i8, int i9, 
                              int i10, int i11, int i12) {     
    
    // Verbose output
    if (verbose >= 3) cout<<"(VRdisplay): adding dataFile "<<dataFileName
                          <<" to display "<<displayNameSav<<endl;
    
    // Create new dataFilePtr with specified format
    vector<int> format;
    getFormat(format,i1,i2,i3,i4,i5,i6,i7,i8,i9,i10,i11,i12);
    dataFilePtr = new VRfile(format);
    
    // Read file 
    bool ok = dataFilePtr->readFile(dataFileName, verbose);
    if (ok) {
      
      // Protect against memory leaks
      if (hasDataSav) {
        delete dataPtr;
        delete dataNormPtr;
        delete dataTwoSigmaPtr;
      }
        
      // Create new VRgraph to contain data points
      dataPtr = new VRgraphData(dataFilePtr);
      dataNormPtr = new VRgraphNorm(dataPtr,dataPtr);
      dataNormPtr->setColors(0);
      // Copies to display two-sigma uncertainties
      dataTwoSigmaPtr = new VRgraphData(dataFilePtr);
      dataTwoSigmaPtr->setColors(-2);      
      dataTwoSigmaNormPtr = new VRgraphNorm(dataTwoSigmaPtr,dataTwoSigmaPtr);
      dataTwoSigmaNormPtr->setColors(-2);      

      // This display has data
      hasDataSav = true;
      if (showRatioPad) addPad();
      
    } else {
      cout<<"VRDisplay::setDataFile() error reading "<<dataFileName<<endl;
    }
    
    // Return
    return ok;
  }
  
  //------------------------------------------------------------------------
  // addHisFile() : 
  // Add file with saved histogram,
  // with i1 : i12 specifying the data format, columnwise (see VINCIA docs)
  bool VRdisplay::addHisFile(string hisFileName, int i1, int i2, int i3, 
                             int i4, int i5, int i6, int i7, int i8, int i9, 
                             int i10, int i11, int i12) {     
    
    // Verbose output
    if (verbose >= 3) cout<<"(VRdisplay): adding histogram file "<<hisFileName
                          <<" to display "<<displayNameSav<<endl;
    
    // Add to list of hisFilePtrs.
    int nHisFiles = hisFilePtrs.size();
    hisFilePtrs.resize(nHisFiles+1);
    
    // Create new dataFilePtr with specified format
    vector<int> format;
    getFormat(format,i1,i2,i3,i4,i5,i6,i7,i8,i9,i10,i11,i12);
    hisFilePtrs[nHisFiles] = new VRfile(format);
    
    // Read file and return success/failure
    bool ok = hisFilePtrs[nHisFiles]->readFile(hisFileName, verbose);
    if (ok) {
      
      // Add to list of graphsHis
      int nGraphsHis = graphsHis.size();
      graphsHis.resize(nGraphsHis+1);
    
      // Create new VRgraph to contain histogram points
      graphsHis[nGraphsHis] = VRgraphHis( hisFilePtrs[nHisFiles] );     
      graphsHis[nGraphsHis].setColors(graphs.size()+graphsHis.size());

      VRgraphHis* hisPtr = &graphsHis[nGraphsHis];
      if (hasDataSav) {
        graphsHisNorm.push_back( VRgraphNorm(hisPtr,dataPtr) );
      } else if (graphs.size() >= 1) {
        graphsHisNorm.push_back(VRgraphNorm(hisPtr,&graphs[0]));
      } else {
        graphsHisNorm.push_back(VRgraphNorm(hisPtr,&graphsHis[0]));
      }
      graphsHisNorm[nGraphsHis].setColors(graphs.size()+graphsHis.size());
      
      // This display can have a ratio
      if (showRatioPad && nPad <= 1) addPad();
      
    } else {
      cout<<"VRDisplay::setHisFile() error reading "<<hisFileName<<endl;
    }
    
    // Return
    return ok; 
    
  }

  //------------------------------------------------------------------------
  // update(): 
  // Update Display 
  void VRdisplay::update() {
    
    // CD to this display
    displayCanvasPtr->cd();        
    
    // If several pads, first do main plot
    gotoPad(1);
    
    // Update and (re)draw axis
    if (!hasAxes()) {
      if (graphs.size() >= 1) setAxes(graphs[0].getTH1DPtr());
      else return;
    }
    
    // Update max/min if possible.
    if (graphs.size() >= 1) {
      TH1D* hisPtr = graphs[0].getTH1DPtr();
      if (hisPtr != NULL) {
        // Check if min/max changed by user
        double yMinSav = axesPtr->GetMinimum();
        double yMaxSav = axesPtr->GetMaximum();
        double yMinNow = hisPtr->GetMinimum();
        double yMaxNow = hisPtr->GetMaximum();
        if (yMinSav != yMinNow || yMaxSav != yMaxNow) {
          axesPtr->SetMinimum(yMinNow);
          axesPtr->SetMaximum(yMaxNow);
        }
        
        // Also update log Y axis properties (force linear if yMin <= 0.0)
        if ( yMinNow <= 0.0 && gPad->GetLogy() ) setLogY(false); 
        else if ( yMinNow > 0.0 && logY && !gPad->GetLogy() ) setLogY(true);
        
        // Update title 
        setTitle();
      }
    }
    
    // Draw axes
    if (nPad >= 2) {
      // Switch off X axis title for main histo
      axesPtr->GetXaxis()->SetTitle("");
      // Draw relUnc pads
      if (showRelUncPads) {
        int iPad = 1;
        for (int iGraph = 0; iGraph < int(graphs.size()); ++iGraph) {
          if (graphs[iGraph].hasOuter()) {
            ++iPad;
            displayCanvasPtr->cd(iPad);
            axesVarPtr->Draw();
          }
        }
      }
      // Draw ratio pad
      if (showRatioPad && hasDataSav && dataPtr) {
        displayCanvasPtr->cd(nPad);
        axesNormPtr->Draw();        
      }
    }
    // Draw main display
    displayCanvasPtr->cd(1);
    axesPtr->Draw();      
    
    // Update data 
    // (data is always shown with 1 sigma error bars)
    if (hasDataSav && dataPtr) {
      displayCanvasPtr->cd(1);
      dataTwoSigmaPtr->update(normalize,normFac,2.0);
      dataTwoSigmaPtr->DrawOuter("E0Z");
      dataPtr->update(normalize,normFac,1.0);
      dataPtr->DrawOuter("E10");
      dataPtr->Draw("PE1");
      if (dataNormPtr && nPad >= 2 && showRatioPad) {
        displayCanvasPtr->cd(nPad);
        dataTwoSigmaNormPtr->update(true,1.0,2.0);
        dataTwoSigmaNormPtr->DrawOuter("e2");
        dataTwoSigmaNormPtr->Draw("e2");
        dataNormPtr->update(true,1.0,1.0);
        dataNormPtr->DrawOuter("e2");
        dataNormPtr->Draw("e2");
        axesNormPtr->Draw("sameaxis");
        //gPad->RedrawAxis();
        // Draw center line at 1
        xCtrPtr->Draw();
      }
    }
    
    // Update graphs made from saved histograms (dat files)
    displayCanvasPtr->cd(1);
    for (unsigned int iGraph=0; iGraph<graphsHis.size(); ++iGraph) {
      
      // Set X scatter
      double scatterX = 0.0;
      //0.1*pow(-1.0,(iGraph+graphs.size())*1.0) 
      //  * (iGraph+graphs.size())/(graphsHis.size()+graphs.size());
      
      // Skip if update fails
      if (!graphsHis[iGraph].update(normalize,normFac,sigmaStat,scatterX)) 
        continue; 
      
      // Outer band
      if (graphsHis[iGraph].hasOuter()) {
        int fillStyle = 3354;
        if (iGraph == 1) fillStyle = 3345;
        else if (iGraph >= 2) fillStyle = 3307;
        graphsHis[iGraph].getOuter()->SetFillStyle(fillStyle);
      }
      
      // Draw graph
      graphsHis[iGraph].DrawOuter("20");
      graphsHis[iGraph].Draw("LPZ0");
    }
    
    // Update graphs
    displayCanvasPtr->cd(1);
    for (unsigned int iGraph=0; iGraph<graphs.size(); ++iGraph) {
      
      // Set X scatter
      double scatterX = 0.1*pow(-1.0,iGraph*1.0) * iGraph/graphs.size();
      
      // Skip if update fails
      if (!graphs[iGraph].update(normalize,normFac,sigmaStat,scatterX)) continue; 
      
      // Outer band
      if (graphs[iGraph].hasOuter()) {
        int fillStyle = 3354;
        if (iGraph == 1) fillStyle = 3345;            
        else if (iGraph >= 2) fillStyle = 3307;
        graphs[iGraph].getOuter()->SetFillStyle(fillStyle);                    
      }
      
      // Draw graph, first one with systematic errors as a box
      if (iGraph == 0) graphs[iGraph].DrawOuter("20");    
      else graphs[iGraph].DrawOuter("10");    
      graphs[iGraph].Draw("LPZ0");    
    }
    
    // Update normalized graphs
    bool hasRatioPad = false;
    if (dataNormPtr && nPad >=2 && hasDataSav && showRatioPad) {           
      hasRatioPad = true;
      displayCanvasPtr->cd(nPad);
      if (graphsNorm.size() < graphs.size()) {
        graphsNorm.resize(0);
        for (unsigned int iGraph=0; iGraph<graphs.size(); ++iGraph) {
          graphsNorm.push_back(VRgraphNorm(&graphs[iGraph],dataPtr));  
          graphsNorm[iGraph].setColors(&graphs[iGraph]);
          if (graphsNorm[iGraph].hasOuter()) {
            int fillStyle = 3354;
            if (iGraph == 1) fillStyle = 3345; 
            else if (iGraph >= 2) fillStyle = 3307;
            graphsNorm[iGraph].SetFillStyle(fillStyle);            
          }
        }
      }
      // Draw saved histo ratio graphs
      for (unsigned int iGraph=0; iGraph<graphsHis.size(); ++iGraph) {
        // Update
        graphsHisNorm[iGraph].update(sigmaStat);
        // Skip if no points to be plotted
        if (graphsHisNorm[iGraph].GetN() <= 0) continue;
        // Draw graph
        graphsHisNorm[iGraph].DrawOuter("20");
        graphsHisNorm[iGraph].Draw("LPZ0");
      }
      // Draw MC ratio graphs
      for (unsigned int iGraph=0; iGraph<graphs.size(); ++iGraph) {
        // Update
        graphsNorm[iGraph].update(sigmaStat);
        // Skip if no points to be plotted
        if (graphsNorm[iGraph].GetN() <= 0) continue;
        graphsNorm[iGraph].DrawOuter("20");
        graphsNorm[iGraph].Draw("LPZ0");
      }
      // Redraw upper boundary of ratio plot in case hidden by fill styles
      xUppPtr->Draw();
    }
    
    // Update relative uncertainty variation pads
    int nUncVarPads = nPad - hasRatioPad -1;
    if (nUncVarPads >= 1) {
      int iPad = 1;
      for (int iGraph = 0; iGraph < int(graphs.size()); ++iGraph) {
        // Only do RelUnc pads for graphs that have (>1) outer bands
        if (!graphs[iGraph].hasOuter() || graphs[iGraph].nOuter() <= 1) 
          continue;
        ++iPad;
        // Sanity check: allow for max 3 separate RelUnc pads
        if (iPad > nPad || iPad >= 3) continue;
        displayCanvasPtr->cd(iPad);
        // Create vector if not already done
        if (graphsVar.size() <= 1) {
          graphsVar.resize(0);
          for (int iVar=1; iVar<graphs[iGraph].nOuter(); ++iVar) {
            graphsVar.push_back(VRgraphVar(&graphs[iGraph],iVar));
            graphsVar[graphsVar.size()-1].setStyle(iVar);
          }
        }
        for (int iVar=0; iVar<int(graphsVar.size()); ++iVar) {
          graphsVar[iVar].update();
          // Skip if no points to be plotted
          if (graphsVar[iVar].GetN() <= 0) continue;
          // Draw area-filled graphs 
          if (graphsVar[iVar].GetFillStyle() != 0) 
            graphsVar[iVar].Draw("e3line");
          // Draw line graphs
          else {
            graphsVar[iVar].Draw("Xline");
          }
        }
      }
    }
    
    // Draw objects 
    // 1: lines
    displayCanvasPtr->cd(1);
    if (linesPtr.size() >= 1) {
      for (unsigned int iLine=0; iLine < linesPtr.size(); ++iLine)           
        linesPtr[iLine]->Draw();      
    }
    
    // Draw title
    displayCanvasPtr->cd(1);      
    if (hasTitle) titleTexPtr->Draw();
    
    // Draw MC names
    displayCanvasPtr->cd(1);
    if (!hasMCnames) setMCnames(settingsPtr->MCnames);
    if (hasMCnames) MCnamesPtr->Draw();
    
    // Draw data ref
    displayCanvasPtr->cd(1);
    if (hasDataSav) {
      if (!hasDataRef) setDataRef(dataFilePtr->getREF());
      if (hasDataRef) dataRefPtr->Draw();
    }
    
    // Draw legend (so it appears on top of curves)
    displayCanvasPtr->cd(1);
    if (!hasLegend) setLegend(settingsPtr->legend);
    if (hasLegend) legendPtr->Draw();
    
    // Draw signature
    if (!hasSignature) {
      double xR = 1.-0.9*gPad->GetRightMargin();      
      double yR = gPad->GetBottomMargin();
      if (nPad >= 2) yR = 0.;
      signaturePtr = new TLatex(xR,yR,signature.c_str());
      signaturePtr->SetNDC(true);
      signaturePtr->SetTextAngle(90);
      signaturePtr->SetTextAlign(13);
      signaturePtr->SetTextColor(12);
      signaturePtr->SetTextFont(42);
      double sigSize = 0.02 + 0.003*nPad;
      signaturePtr->SetTextSize(sigSize);
      hasSignature = true;
    }
    signaturePtr->Draw();
    
    // Then update canvas
    displayCanvasPtr->cd();
    displayCanvasPtr->Update();      
  }
  
  //------------------------------------------------------------------------
  // Save display image
  void VRdisplay::saveDisplay(string prefix,  
                              string fileType, string option) {
    // Default file name is the name of display
    string fileName = displayNameSav;
    // If specific prefix given, append display name to that 
    if (prefix != "none") fileName = prefix+"-"+displayNameSav;          
    // First loop for specific file types
    if (fileType.find("eps") != string::npos) fileName+=".eps";
    else if (fileType.find("ps") != string::npos) fileName+=".ps";
    else if (fileType.find("gif") != string::npos) fileName+=".gif";
    else if (fileType.find("jpg") != string::npos) fileName+=".jpg";
    else if (fileType.find("png") != string::npos) fileName+=".png";
    else if (fileType.find("pdf") != string::npos) fileName+=".pdf";
    // Then try generic extension. Don't add dot if ext already has one.
    else if (fileType.find(".") != string::npos) fileName+=fileType;
    else fileName+="."+fileType;
    displayCanvasPtr->SaveAs(fileName.c_str(),option.c_str());
  }    
  
  //------------------------------------------------------------------------
  // saveHistogram():
  bool VRdisplay::saveHistogram(string prefix, string ext) {
    
    // Default file name is the name of display
    string fileName = displayNameSav;
    // If specific prefix given, append display name to that 
    if (prefix != "none") fileName = prefix+"-"+displayNameSav;          
    fileName += "."+ext;
    FILE * pFile = fopen (fileName.c_str(), "w");
    if (pFile == NULL) {
      perror ("(VRdisplay::saveHistogram): Error opening file");
      return false;
    }
    else {
      // Number of bins to print
      TH1D* hisPtr = getTH1DPtr();
      int nBin = hisPtr->GetNbinsX();
      for (int iBin=1; iBin<=nBin; iBin++){
        double dx  = hisPtr->GetBinWidth(iBin);
        double xLo = hisPtr->GetBinLowEdge(iBin);
        if ( abs(xLo/dx) < 1e-12 ) xLo = 0.;
        double xHi = xLo + dx;
        if ( abs(xHi/dx) < 1e-12 ) xHi = 0.;
        double xC  = xLo + (xHi - xLo)/2.;
        double y   = hisPtr->GetBinContent(iBin);
        // Output standard 1-sigma uncertainties
        double ey  = hisPtr->GetBinError(iBin);
        // Systematic errors ?
        double eyThHi(0.), eyThLo(0.);
        if (graphs[0].hasOuter()) {
          for (int iVar=0; iVar < graphs[0].nOuter(); ++iVar) {
            eyThHi = max(eyThHi,getTH1DPtr(0,iVar)->GetBinContent(iBin) - y);
            eyThLo = max(eyThLo,y - getTH1DPtr(0,iVar)->GetBinContent(iBin));
          }
        }
        
        // If has systematic uncertainties
        if (graphs[0].hasOuter()) {
          fprintf(pFile,
                  "%14.6g %14.6g %14.6g %14.6g %14.6g %14.6g %14.6g %14.6g \n",
                  xLo,xC,xHi,y,ey,ey,eyThHi,eyThLo); 
        } else {
          // Default: write out stat error only
          fprintf(pFile,"%14.6g %14.6g %14.6g %14.6g %14.6g %14.6g \n",
                  xLo,xC,xHi,y,ey,ey); 
          
        }
      }
      
      // Write format as text comment
      if (graphs[0].hasOuter()) {
        fprintf(pFile,"#          xLo             xC            xHi              y        +dyStat        -dyStat          +dyTh          -dyTh\n");
      } 
      else {
        fprintf(pFile,"#          xLo             xC            xHi              y        +dyStat        -dyStat\n");
      }
      
      fprintf(pFile,"# Info:\nTIT = %s\n",hisPtr->GetTitle());
      if (hasLegend && legendSav != "") {
        fprintf(pFile,"SRC = %s\n",legendSav.c_str());
      }
      if (hasMCnames && MCnamesSav != "") {
        fprintf(pFile,"REF = %s\n",MCnamesSav.c_str());
      }
      fprintf(pFile,"#\n# File printed by VinciaRoot\n");        
      fprintf(pFile,"# VinciaRoot is distributed as part of the VINCIA code\n"); 
      fprintf(pFile,"# http://projects.hepforge.org/vincia/\n"); 
      fprintf(pFile,"# http://arxiv.org/abs/arXiv:1102.2126\n");
      fclose(pFile);
      return true;
    }
    
  }
  
  //========================================================================
  // VRdisplay: Plot titles, legends, descriptions
  
  //------------------------------------------------------------------------
  // setTitle():
  // Set title
  void VRdisplay::setTitle() {
    // Can only define title if histogram has been defined
    if (!hasAxes() || graphs.size() == 0) return;
    // Get current histogram title from histogram pointer
    TH1D* hisPtr = graphs[0].getTH1DPtr();
    string hisTitle = hisPtr->GetTitle();
    // Destroy pointer if title already exists (to avoid mem leak)
    if (hasTitle) delete titleTexPtr;
    // Get axis definitions from axis histogram
    int NBins = axesPtr->GetNbinsX();
    double xLo = axesPtr->GetBinLowEdge(1);
    double xHi = axesPtr->GetBinLowEdge(NBins+1);
    double xTit = xLo + (xHi-xLo)/2.;
    double yMin = axesPtr->GetMinimum();
    double yMax = axesPtr->GetMaximum();
    double dyTit = 0.07;
    double yTit = yMin + (1.-dyTit)*(yMax-yMin);
    if (gPad->GetLogy()) {
      double LnyC = log(yMin) + (1.-dyTit)*log(yMax/yMin);
      yTit = exp(LnyC);
    }      
    // Create new title tex pointer
    titleTexPtr = new TLatex(xTit,yTit,hisTitle.c_str());
    hasTitle    = true;
    titleTexPtr->SetTextFont(62);
    titleTexPtr->SetTextAlign(22);
    double titSize = (1.05+0.025*(nPad-1))*titleTexPtr->GetTextSize();
    titleTexPtr->SetTextSize(titSize);
    titleTexPtr->SetTextColor(1);
  }

  //------------------------------------------------------------------------
  // Set MC name title
  void VRdisplay::setMCnames(string MCnamesIn) {
    if (!hasAxes() || graphs.size() == 0) return;
    // Destroy pointer if MCnames already exists (to avoid mem leak)
    if (hasMCnames) delete MCnamesPtr;
    // Get axis definitions from axis histogram
    int NBins = axesPtr->GetNbinsX();
    double xLo = axesPtr->GetBinLowEdge(1);
    double xHi = axesPtr->GetBinLowEdge(NBins+1);
    double xTit = xLo + (xHi-xLo)/2.;
    double yMin = axesPtr->GetMinimum();
    double yMax = axesPtr->GetMaximum();
    double dyTit = 0.085;
    double yTit = yMin + dyTit*(yMax-yMin);
    if (gPad->GetLogy()) {
      double LnyC = log(yMin) + dyTit*log(yMax/yMin);
      yTit = exp(LnyC);
    }      
    MCnamesPtr = new TLatex(xTit,yTit,MCnamesIn.c_str());
    MCnamesPtr->SetTextFont(42);
    MCnamesPtr->SetTextAlign(23);
    MCnamesPtr->SetTextColor(12);
    double namesSize = MCnamesPtr->GetTextSize();
    double scaleFac  = 1. + (nPad-1)/10.;
    MCnamesPtr->SetTextSize(0.7*scaleFac*namesSize);
    hasMCnames = true;
    MCnamesSav = MCnamesIn;
  }
  
  //------------------------------------------------------------------------
  // Set Data reference title
  void VRdisplay::setDataRef(string dataRefIn) {
    if (!hasAxes() || graphs.size() == 0) return;
    // Destroy pointer if MCnames already exists (to avoid mem leak)
    if (hasDataRef) delete dataRefPtr;
    if (dataRefIn == "none") return;
    // Get axis definitions from axis histogram
    int NBins = axesPtr->GetNbinsX();
    double xLo = axesPtr->GetBinLowEdge(1);
    double xHi = axesPtr->GetBinLowEdge(NBins+1);
    double xTit = xLo + (xHi-xLo)/2.;
    double yMin = axesPtr->GetMinimum();
    double yMax = axesPtr->GetMaximum();
    double dyTit = 0.112;
    double yTit = yMin + dyTit*(yMax-yMin);
    if (gPad->GetLogy()) {
      double LnyC = log(yMin) + dyTit*log(yMax/yMin);
      yTit = exp(LnyC);
    }
    // Add "Data from ..." if missing
    if (dataRefIn.find("Data") == string::npos 
        && dataRefIn.find("data") == string::npos) 
      dataRefIn = "Data from " + dataRefIn;      
    dataRefPtr = new TLatex(xTit,yTit,dataRefIn.c_str());
    dataRefPtr->SetTextFont(42);
    dataRefPtr->SetTextAlign(21);
    dataRefPtr->SetTextColor(12);
    double namesSize = dataRefPtr->GetTextSize();
    double scaleFac  = 1. + (nPad-1)/10.;
    dataRefPtr->SetTextSize(0.7*scaleFac*namesSize);
    hasDataRef = true;
  }
  
  //------------------------------------------------------------------------
  // Set Legend (align: -1 (auto), 0: left, 1:right)
  void VRdisplay::setLegend(vector<string> legendIn, int alignIn) {
    if (!hasAxes() || graphs.size() == 0 || legendIn.size() == 0) return;
    // Destroy pointer if MCnames already exists (to avoid mem leak)
    if (hasLegend) delete legendPtr;
    // Find relUnc of axis
    int axisBins   = axesPtr->GetNbinsX();
    double axisMin = axesPtr->GetBinLowEdge(1);
    double axisCtr = 0.5*(axesPtr->GetBinLowEdge(axisBins+1) + axisMin);
    double xMax = axisMin;
    double yMax = -1e9;
    // Check if legend should be placed on left or right
    if (alignIn == 0) leftLegend = true;
    else if (alignIn == 1) leftLegend = false;
    else {
      if (hasDataSav && dataPtr->GetN() >= 1) {
        int nPoints = dataPtr->GetN();
        double xTmp,yTmp;
        for (int iPoint = 0; iPoint < nPoints; iPoint++) {          
          dataPtr->GetPoint(iPoint,xTmp,yTmp);
          if (yTmp > yMax) {
            yMax = yTmp;
            xMax = xTmp;
          }
        }
      }
      if (graphs.size() >= 1) {
        for (int iGraph=0; iGraph < int(graphs.size()); ++iGraph) {
          int nPoints = graphs[iGraph].GetN();
          double xTmp,yTmp;
          for (int iPoint = 0; iPoint < nPoints; iPoint++) {          
            graphs[0].GetPoint(iPoint,xTmp,yTmp);
            if (yTmp > yMax) {
              yMax = yTmp;
              xMax = xTmp;
            }
          }
        }
      }
      // Set legend position Left or Right
      if ( xMax >= axisCtr ) leftLegend = true;
      else leftLegend = false;      
    }
    
    // Legends use absolute coordinates (sucks)
    double xLeg = 0.56;
    if (leftLegend) xLeg = 0.18;
    double xLegEnd = xLeg + 0.33;
    double dy = 0.069;
    double yLegTop = 0.82;
    double yLegBot = yLegTop - dy*(legendIn.size()+hasDataSav+graphsHis.size());
    legendPtr = new TLegend(xLeg,yLegBot,xLegEnd,yLegTop,NULL,"rNDC");
    hasLegend = true;
    legendPtr->SetBorderSize(0);
    legendPtr->SetFillStyle(0);
    // Add data legend
    if (hasDataSav) {
      string datLeg = dataFilePtr->getSRC();
      if (datLeg == "" || datLeg == "none") datLeg = "Data";
      TLegendEntry *entry=legendPtr->AddEntry("dat",datLeg.c_str(),"p");
      entry->SetMarkerStyle(dataPtr->GetMarkerStyle());
      entry->SetMarkerSize(dataPtr->GetMarkerSize());
      entry->SetMarkerColor(dataPtr->GetMarkerColor());
      entry->SetTextAlign(12);
      entry->SetTextFont(42);
      entry->SetTextColor(1);	
    }
    // Add histogram legends
    for (int iLeg = 0; iLeg < min(int(graphs.size()), int(legendIn.size()));
         ++iLeg) {
      TLegendEntry *entry=legendPtr->AddEntry("mc",legendIn[iLeg].c_str(),"lP");
      entry->SetLineColor(graphs[iLeg].GetLineColor());
      entry->SetLineStyle(graphs[iLeg].GetLineStyle());
      entry->SetLineWidth(graphs[iLeg].GetLineWidth());
      entry->SetMarkerStyle(graphs[iLeg].GetMarkerStyle());
      entry->SetMarkerSize(graphs[iLeg].GetMarkerSize());
      entry->SetMarkerColor(graphs[iLeg].GetMarkerColor());
      entry->SetTextAlign(12);
      entry->SetTextFont(42);
      entry->SetTextColor(1);
    }
    // Add legends from saved histograms
    for (int iLeg = 0; iLeg < min(int(graphsHis.size()), int(legendIn.size()));
         ++iLeg) {
      // Safety catch
      if (iLeg >= int(hisFilePtrs.size())) continue;
      string hisLeg = hisFilePtrs[iLeg]->getSRC();
      if (hisLeg == "" || hisLeg == "none") hisLeg = "Data";
      TLegendEntry *entry=legendPtr->AddEntry("his",hisLeg.c_str(),"lP");
      entry->SetMarkerStyle(graphsHis[iLeg].GetMarkerStyle());
      entry->SetMarkerSize(graphsHis[iLeg].GetMarkerSize());
      entry->SetMarkerColor(graphsHis[iLeg].GetMarkerColor());
      entry->SetLineColor(graphsHis[iLeg].GetLineColor());
      entry->SetLineStyle(graphsHis[iLeg].GetLineStyle());
      entry->SetLineWidth(graphsHis[iLeg].GetLineWidth());
      entry->SetTextAlign(12);
      entry->SetTextFont(42);
      entry->SetTextColor(1);	
    }
    // Save legend of main (=first) histogram for use by saveHistogram().      
    legendSav = (legendIn.size() > 0) ? legendIn[0] : "";
  }
    
} // End namespace Vincia

